from __future__ import annotations

from amulet.core.palette.biome_palette import BiomePalette
from amulet.core.palette.block_palette import BlockPalette

from . import biome_palette, block_palette

__all__ = ["BiomePalette", "BlockPalette", "biome_palette", "block_palette"]
